@@ -2,11 +2,15 @@ use cap_primitives::ambient_authority;
 #[cfg(not(windows))]
 use cap_primitives::fs::symlink;
 use cap_primitives::fs::{open_dir_nofollow, set_times, set_times_nofollow};
+#[cfg(all(windows, feature = "async_std", feature = "fs_utf8"))]
+use cap_primitives::fs::{stat, FollowSymlinks};
 #[cfg(windows)]
 use cap_primitives::fs::{symlink_dir, symlink_file};
 use io_lifetimes::AsFilelike;
 use std::io;
 use std::path::Path;
+#[cfg(feature = "async_std")]
+use {async_std::task::spawn_blocking, async_trait::async_trait};
 
 pub use cap_primitives::fs::SystemTimeSpec;
 
@@ -92,6 +96,127 @@ pub trait DirExt {
     fn remove_file_or_symlink<P: AsRef<Path>>(&self, path: P) -> io::Result<()>;
 }
 
+/// Extension trait for `Dir`, async.
+///
+/// The path parameters include `Send` for the `async_trait` macro.
+#[cfg(feature = "async_std")]
+#[async_trait]
+pub trait AsyncDirExt {
+    /// Set the last access time for a file on a filesystem.
+    ///
+    /// This corresponds to [`filetime::set_file_atime`].
+    ///
+    /// [`filetime::set_file_atime`]: https://docs.rs/filetime/latest/filetime/fn.set_file_atime.html
+    async fn set_atime<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+        atime: SystemTimeSpec,
+    ) -> io::Result<()>;
+
+    /// Set the last modification time for a file on a filesystem.
+    ///
+    /// This corresponds to [`filetime::set_file_mtime`].
+    ///
+    /// [`filetime::set_file_mtime`]: https://docs.rs/filetime/latest/filetime/fn.set_file_mtime.html
+    async fn set_mtime<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+        mtime: SystemTimeSpec,
+    ) -> io::Result<()>;
+
+    /// Set the last access and modification times for a file on a filesystem.
+    ///
+    /// This corresponds to [`filetime::set_file_times`].
+    ///
+    /// [`filetime::set_file_times`]: https://docs.rs/filetime/latest/filetime/fn.set_file_times.html
+    async fn set_times<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+        atime: Option<SystemTimeSpec>,
+        mtime: Option<SystemTimeSpec>,
+    ) -> io::Result<()>;
+
+    /// Set the last access and modification times for a file on a filesystem.
+    /// This function does not follow symlink.
+    ///
+    /// This corresponds to [`filetime::set_symlink_file_times`].
+    ///
+    /// [`filetime::set_symlink_file_times`]: https://docs.rs/filetime/latest/filetime/fn.set_symlink_file_times.html
+    async fn set_symlink_times<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+        atime: Option<SystemTimeSpec>,
+        mtime: Option<SystemTimeSpec>,
+    ) -> io::Result<()>;
+
+    /// Creates a new symbolic link on a filesystem.
+    ///
+    /// This corresponds to [`std::os::unix::fs::symlink`], except that
+    /// it's supported on non-Unix platforms as well, and it's not guaranteed
+    /// to be atomic.
+    ///
+    /// [`std::os::unix::fs::symlink`]: https://doc.rust-lang.org/std/os/unix/fs/fn.symlink.html
+    async fn symlink<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()>;
+
+    /// Creates a new file symbolic link on a filesystem.
+    ///
+    /// This corresponds to [`std::os::windows::fs::symlink_file`], except that
+    /// it's supported on non-Windows platforms as well, and it's not
+    /// guaranteed to fail if the target is not a file.
+    ///
+    /// [`std::os::windows::fs::symlink_file`]: https://doc.rust-lang.org/std/os/windows/fs/fn.symlink_file.html
+    async fn symlink_file<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()>;
+
+    /// Creates a new directory symbolic link on a filesystem.
+    ///
+    /// This corresponds to [`std::os::windows::fs::symlink_dir`], except that
+    /// it's supported on non-Windows platforms as well, and it's not
+    /// guaranteed to fail if the target is not a directory.
+    ///
+    /// [`std::os::windows::fs::symlink_dir`]: https://doc.rust-lang.org/std/os/windows/fs/fn.symlink_dir.html
+    async fn symlink_dir<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()>;
+
+    /// Similar to `cap_std::fs::Dir::open_dir`, but fails if the path names a
+    /// symlink.
+    async fn open_dir_nofollow<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+    ) -> io::Result<Self>
+    where
+        Self: Sized;
+
+    /// Removes a file or symlink from a filesystem.
+    ///
+    /// Removal of symlinks has different behavior under Windows - if a symlink
+    /// points to a directory, it cannot be removed with the `remove_file`
+    /// operation. This method will remove files and all symlinks.
+    async fn remove_file_or_symlink<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+    ) -> io::Result<()>;
+}
+
 /// `fs_utf8` version of `DirExt`.
 #[cfg(all(any(feature = "std", feature = "async_std"), feature = "fs_utf8"))]
 pub trait DirExtUtf8 {
@@ -175,6 +300,112 @@ pub trait DirExtUtf8 {
     fn remove_file_or_symlink<P: AsRef<str>>(&self, path: P) -> io::Result<()>;
 }
 
+/// `fs_utf8` version of `DirExt`.
+///
+/// The path parameters include `Send` for the `async_trait` macro.
+#[cfg(all(feature = "async_std", feature = "fs_utf8"))]
+#[async_trait]
+pub trait AsyncDirExtUtf8 {
+    /// Set the last access time for a file on a filesystem.
+    ///
+    /// This corresponds to [`filetime::set_file_atime`].
+    ///
+    /// [`filetime::set_file_atime`]: https://docs.rs/filetime/latest/filetime/fn.set_file_atime.html
+    async fn set_atime<P: AsRef<str> + Send>(
+        &self,
+        path: P,
+        atime: SystemTimeSpec,
+    ) -> io::Result<()>;
+
+    /// Set the last modification time for a file on a filesystem.
+    ///
+    /// This corresponds to [`filetime::set_file_mtime`].
+    ///
+    /// [`filetime::set_file_mtime`]: https://docs.rs/filetime/latest/filetime/fn.set_file_mtime.html
+    async fn set_mtime<P: AsRef<str> + Send>(
+        &self,
+        path: P,
+        mtime: SystemTimeSpec,
+    ) -> io::Result<()>;
+
+    /// Set the last access and modification times for a file on a filesystem.
+    ///
+    /// This corresponds to [`filetime::set_file_times`].
+    ///
+    /// [`filetime::set_file_times`]: https://docs.rs/filetime/latest/filetime/fn.set_file_times.html
+    async fn set_times<P: AsRef<str> + Send>(
+        &self,
+        path: P,
+        atime: Option<SystemTimeSpec>,
+        mtime: Option<SystemTimeSpec>,
+    ) -> io::Result<()>;
+
+    /// Set the last access and modification times for a file on a filesystem.
+    /// This function does not follow symlink.
+    ///
+    /// This corresponds to [`filetime::set_symlink_file_times`].
+    ///
+    /// [`filetime::set_symlink_file_times`]: https://docs.rs/filetime/latest/filetime/fn.set_symlink_file_times.html
+    async fn set_symlink_times<P: AsRef<str> + Send>(
+        &self,
+        path: P,
+        atime: Option<SystemTimeSpec>,
+        mtime: Option<SystemTimeSpec>,
+    ) -> io::Result<()>;
+
+    /// Creates a new symbolic link on a filesystem.
+    ///
+    /// This corresponds to [`std::os::unix::fs::symlink`], except that
+    /// it's supported on non-Unix platforms as well, and it's not guaranteed
+    /// to be atomic.
+    ///
+    /// [`std::os::unix::fs::symlink`]: https://doc.rust-lang.org/std/os/unix/fs/fn.symlink.html
+    async fn symlink<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()>;
+
+    /// Creates a new file symbolic link on a filesystem.
+    ///
+    /// This corresponds to [`std::os::windows::fs::symlink_file`], except that
+    /// it's supported on non-Windows platforms as well, and it's not
+    /// guaranteed to fail if the target is not a file.
+    ///
+    /// [`std::os::windows::fs::symlink_file`]: https://doc.rust-lang.org/std/os/windows/fs/fn.symlink_file.html
+    async fn symlink_file<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()>;
+
+    /// Creates a new directory symbolic link on a filesystem.
+    ///
+    /// This corresponds to [`std::os::windows::fs::symlink_dir`], except that
+    /// it's supported on non-Windows platforms as well, and it's not
+    /// guaranteed to fail if the target is not a directory.
+    ///
+    /// [`std::os::windows::fs::symlink_dir`]: https://doc.rust-lang.org/std/os/windows/fs/fn.symlink_dir.html
+    async fn symlink_dir<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()>;
+
+    /// Similar to `cap_std::fs::Dir::open_dir`, but fails if the path names a
+    /// symlink.
+    async fn open_dir_nofollow<P: AsRef<str> + Send>(&self, path: P) -> io::Result<Self>
+    where
+        Self: Sized;
+
+    /// Removes a file or symlink from a filesystem.
+    ///
+    /// Removal of symlinks has different behavior under Windows - if a symlink
+    /// points to a directory, it cannot be removed with the `remove_file`
+    /// operation. This method will remove files and all symlinks.
+    async fn remove_file_or_symlink<P: AsRef<str> + Send>(&self, path: P) -> io::Result<()>;
+}
+
 #[cfg(feature = "std")]
 impl DirExt for cap_std::fs::Dir {
     #[inline]
@@ -330,157 +561,296 @@ impl DirExt for cap_std::fs::Dir {
 }
 
 #[cfg(feature = "async_std")]
-impl DirExt for cap_async_std::fs::Dir {
+#[async_trait]
+impl AsyncDirExt for cap_async_std::fs::Dir {
     #[inline]
-    fn set_atime<P: AsRef<Path>>(&self, path: P, atime: SystemTimeSpec) -> io::Result<()> {
-        set_times(
-            &self.as_filelike_view::<std::fs::File>(),
-            path.as_ref(),
-            Some(atime),
-            None,
-        )
+    async fn set_atime<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+        atime: SystemTimeSpec,
+    ) -> io::Result<()> {
+        let path = path.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times(
+                &clone.as_filelike_view::<std::fs::File>(),
+                path.as_ref(),
+                Some(atime),
+                None,
+            )
+        })
+        .await
     }
 
     #[inline]
-    fn set_mtime<P: AsRef<Path>>(&self, path: P, mtime: SystemTimeSpec) -> io::Result<()> {
-        set_times(
-            &self.as_filelike_view::<std::fs::File>(),
-            path.as_ref(),
-            None,
-            Some(mtime),
-        )
+    async fn set_mtime<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+        mtime: SystemTimeSpec,
+    ) -> io::Result<()> {
+        let path = path.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times(
+                &clone.as_filelike_view::<std::fs::File>(),
+                path.as_ref(),
+                None,
+                Some(mtime),
+            )
+        })
+        .await
     }
 
     #[inline]
-    fn set_times<P: AsRef<Path>>(
+    async fn set_times<P: AsRef<async_std::path::Path> + Send>(
         &self,
         path: P,
         atime: Option<SystemTimeSpec>,
         mtime: Option<SystemTimeSpec>,
     ) -> io::Result<()> {
-        set_times(
-            &self.as_filelike_view::<std::fs::File>(),
-            path.as_ref(),
-            atime,
-            mtime,
-        )
+        let path = path.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times(
+                &clone.as_filelike_view::<std::fs::File>(),
+                path.as_ref(),
+                atime,
+                mtime,
+            )
+        })
+        .await
     }
 
     #[inline]
-    fn set_symlink_times<P: AsRef<Path>>(
+    async fn set_symlink_times<P: AsRef<async_std::path::Path> + Send>(
         &self,
         path: P,
         atime: Option<SystemTimeSpec>,
         mtime: Option<SystemTimeSpec>,
     ) -> io::Result<()> {
-        set_times_nofollow(
-            &self.as_filelike_view::<std::fs::File>(),
-            path.as_ref(),
-            atime,
-            mtime,
-        )
+        let path = path.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times_nofollow(
+                &clone.as_filelike_view::<std::fs::File>(),
+                path.as_ref(),
+                atime,
+                mtime,
+            )
+        })
+        .await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn symlink<P: AsRef<Path>, Q: AsRef<Path>>(&self, src: P, dst: Q) -> io::Result<()> {
-        symlink(
-            src.as_ref(),
-            &self.as_filelike_view::<std::fs::File>(),
-            dst.as_ref(),
-        )
+    async fn symlink<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        let src = src.as_ref().to_path_buf();
+        let dst = dst.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            symlink(
+                src.as_ref(),
+                &clone.as_filelike_view::<std::fs::File>(),
+                dst.as_ref(),
+            )
+        })
+        .await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn symlink_file<P: AsRef<Path>, Q: AsRef<Path>>(&self, src: P, dst: Q) -> io::Result<()> {
-        self.symlink(src.as_ref(), dst.as_ref())
+    async fn symlink_file<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        let src = src.as_ref().to_path_buf();
+        let dst = dst.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            symlink(
+                src.as_ref(),
+                &clone.as_filelike_view::<std::fs::File>(),
+                dst.as_ref(),
+            )
+        })
+        .await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn symlink_dir<P: AsRef<Path>, Q: AsRef<Path>>(&self, src: P, dst: Q) -> io::Result<()> {
-        self.symlink(src.as_ref(), dst.as_ref())
+    async fn symlink_dir<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        let src = src.as_ref().to_path_buf();
+        let dst = dst.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            symlink(
+                src.as_ref(),
+                &clone.as_filelike_view::<std::fs::File>(),
+                dst.as_ref(),
+            )
+        })
+        .await
     }
 
     #[cfg(windows)]
     #[inline]
-    fn symlink<P: AsRef<Path>, Q: AsRef<Path>>(&self, src: P, dst: Q) -> io::Result<()> {
-        if self.metadata(src.as_ref())?.is_dir() {
-            self.symlink_dir(src.as_ref(), dst.as_ref())
+    async fn symlink<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        let src = src.as_ref().to_path_buf();
+        let dst = dst.as_ref().to_path_buf();
+        let clone = self.clone();
+        if self.metadata(&src).await?.is_dir() {
+            spawn_blocking(move || {
+                symlink_dir(
+                    src.as_ref(),
+                    &clone.as_filelike_view::<std::fs::File>(),
+                    dst.as_ref(),
+                )
+            })
+            .await
         } else {
-            self.symlink_file(src.as_ref(), dst.as_ref())
+            spawn_blocking(move || {
+                symlink_file(
+                    src.as_ref(),
+                    &clone.as_filelike_view::<std::fs::File>(),
+                    dst.as_ref(),
+                )
+            })
+            .await
         }
     }
 
     #[cfg(windows)]
     #[inline]
-    fn symlink_file<P: AsRef<Path>, Q: AsRef<Path>>(&self, src: P, dst: Q) -> io::Result<()> {
-        symlink_file(
-            src.as_ref(),
-            &self.as_filelike_view::<std::fs::File>(),
-            dst.as_ref(),
-        )
+    async fn symlink_file<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        let src = src.as_ref().to_path_buf();
+        let dst = dst.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            symlink_file(
+                src.as_ref(),
+                &clone.as_filelike_view::<std::fs::File>(),
+                dst.as_ref(),
+            )
+        })
+        .await
     }
 
     #[cfg(windows)]
     #[inline]
-    fn symlink_dir<P: AsRef<Path>, Q: AsRef<Path>>(&self, src: P, dst: Q) -> io::Result<()> {
-        symlink_dir(
-            src.as_ref(),
-            &self.as_filelike_view::<std::fs::File>(),
-            dst.as_ref(),
-        )
+    async fn symlink_dir<
+        P: AsRef<async_std::path::Path> + Send,
+        Q: AsRef<async_std::path::Path> + Send,
+    >(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        let src = src.as_ref().to_path_buf();
+        let dst = dst.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            symlink_dir(
+                src.as_ref(),
+                &clone.as_filelike_view::<std::fs::File>(),
+                dst.as_ref(),
+            )
+        })
+        .await
     }
 
     #[inline]
-    fn open_dir_nofollow<P: AsRef<Path>>(&self, path: P) -> io::Result<Self> {
-        match open_dir_nofollow(&self.as_filelike_view::<std::fs::File>(), path.as_ref()) {
-            Ok(file) => Ok(Self::from_std_file(file.into(), ambient_authority())),
-            Err(e) => Err(e),
-        }
+    async fn open_dir_nofollow<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+    ) -> io::Result<Self> {
+        let path = path.as_ref().to_path_buf();
+        let clone = self.clone();
+        spawn_blocking(move || {
+            match open_dir_nofollow(&clone.as_filelike_view::<std::fs::File>(), path.as_ref()) {
+                Ok(file) => Ok(Self::from_std_file(file.into(), ambient_authority())),
+                Err(e) => Err(e),
+            }
+        })
+        .await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn remove_file_or_symlink<P: AsRef<Path>>(&self, path: P) -> io::Result<()> {
-        self.remove_file(path.as_ref())
+    async fn remove_file_or_symlink<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+    ) -> io::Result<()> {
+        self.remove_file(path).await
     }
 
     #[cfg(windows)]
     #[inline]
-    fn remove_file_or_symlink<P: AsRef<Path>>(&self, path: P) -> io::Result<()> {
-        // This operation may race, because it checks the metadata before deleting
-        // the symlink. We tried to do this atomically by ReOpenFile with
-        // DELETE_ON_CLOSE but could not get it to work.
-        fn delete_symlink_to_dir(dir: &cap_async_std::fs::Dir, path: &Path) -> io::Result<()> {
-            use crate::{FollowSymlinks, OpenOptionsFollowExt};
-            use cap_std::fs::OpenOptions;
-            use std::os::windows::fs::OpenOptionsExt;
-            use winapi::um::winbase::{FILE_FLAG_BACKUP_SEMANTICS, FILE_FLAG_OPEN_REPARSE_POINT};
+    async fn remove_file_or_symlink<P: AsRef<async_std::path::Path> + Send>(
+        &self,
+        path: P,
+    ) -> io::Result<()> {
+        use crate::{FollowSymlinks, OpenOptionsFollowExt};
+        use cap_primitives::fs::_WindowsByHandle;
+        use cap_std::fs::OpenOptions;
+        use std::os::windows::fs::OpenOptionsExt;
+        use winapi::um::winbase::{FILE_FLAG_BACKUP_SEMANTICS, FILE_FLAG_OPEN_REPARSE_POINT};
+        use winapi::um::winnt::{DELETE, FILE_ATTRIBUTE_DIRECTORY};
+        let path = path.as_ref();
 
-            let mut opts = OpenOptions::new();
-            opts.read(true);
-            opts.custom_flags(FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS);
-            opts.follow(FollowSymlinks::No);
-            let file = dir.open_with(path, &opts)?;
+        let mut opts = OpenOptions::new();
+        opts.access_mode(DELETE);
+        opts.custom_flags(FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS);
+        opts.follow(FollowSymlinks::No);
+        let file = self.open_with(path, &opts).await?;
 
-            let meta = file.metadata()?;
-            if meta.file_type().is_symlink() {
-                drop(file);
-                // Symlinks that point to directories use the remove_dir, not remove_file,
-                // operation on windows:
-                dir.remove_dir(path)?;
-                Ok(())
-            } else {
-                Err(io::Error::from_raw_os_error(
-                    winapi::shared::winerror::ERROR_DIRECTORY as i32,
-                ))
-            }
+        let meta = file.metadata()?;
+        if meta.file_type().is_symlink()
+            && meta.file_attributes() & FILE_ATTRIBUTE_DIRECTORY == FILE_ATTRIBUTE_DIRECTORY
+        {
+            self.remove_dir(path).await?;
+        } else {
+            self.remove_file(path).await?;
         }
 
-        self.remove_file(path.as_ref())
-            .or_else(|e| delete_symlink_to_dir(&self, path.as_ref()).map_err(|_| e))
+        // Drop the file after calling `remove_file` or `remove_dir`, since
+        // Windows doesn't actually remove the file until after the last open
+        // handle is closed, and this protects us from race conditions where
+        // other processes replace the file out from underneath us.
+        drop(file);
+
+        Ok(())
     }
 }
 
@@ -600,193 +970,268 @@ impl DirExtUtf8 for cap_std::fs_utf8::Dir {
     #[cfg(windows)]
     #[inline]
     fn remove_file_or_symlink<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
-        // This operation may race, because it checks the metadata before deleting
-        // the symlink. We tried to do this atomically by ReOpenFile with
-        // DELETE_ON_CLOSE but could not get it to work.
-        fn delete_symlink_to_dir(dir: &cap_std::fs_utf8::Dir, path: &str) -> io::Result<()> {
-            use crate::{FollowSymlinks, OpenOptionsFollowExt};
-            use cap_std::fs::OpenOptions;
-            use std::os::windows::fs::OpenOptionsExt;
-            use winapi::um::winbase::{FILE_FLAG_BACKUP_SEMANTICS, FILE_FLAG_OPEN_REPARSE_POINT};
+        use crate::{FollowSymlinks, OpenOptionsFollowExt};
+        use cap_primitives::fs::_WindowsByHandle;
+        use cap_std::fs::OpenOptions;
+        use std::os::windows::fs::OpenOptionsExt;
+        use winapi::um::winbase::{FILE_FLAG_BACKUP_SEMANTICS, FILE_FLAG_OPEN_REPARSE_POINT};
+        use winapi::um::winnt::{DELETE, FILE_ATTRIBUTE_DIRECTORY};
+        let path = path.as_ref();
 
-            let mut opts = OpenOptions::new();
-            opts.read(true);
-            opts.custom_flags(FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS);
-            opts.follow(FollowSymlinks::No);
-            let file = dir.open_with(path, &opts)?;
+        let mut opts = OpenOptions::new();
+        opts.access_mode(DELETE);
+        opts.custom_flags(FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS);
+        opts.follow(FollowSymlinks::No);
+        let file = self.open_with(path, &opts)?;
 
-            let meta = file.metadata()?;
-            if meta.file_type().is_symlink() {
-                drop(file);
-                // Symlinks that point to directories use the remove_dir, not remove_file,
-                // operation on windows:
-                dir.remove_dir(path)?;
-                Ok(())
-            } else {
-                Err(io::Error::from_raw_os_error(
-                    winapi::shared::winerror::ERROR_DIRECTORY as i32,
-                ))
-            }
+        let meta = file.metadata()?;
+        if meta.file_type().is_symlink()
+            && meta.file_attributes() & FILE_ATTRIBUTE_DIRECTORY == FILE_ATTRIBUTE_DIRECTORY
+        {
+            self.remove_dir(path)?;
+        } else {
+            self.remove_file(path)?;
         }
 
-        self.remove_file(path.as_ref())
-            .or_else(|e| delete_symlink_to_dir(&self, path.as_ref()).map_err(|_| e))
+        // Drop the file after calling `remove_file` or `remove_dir`, since
+        // Windows doesn't actually remove the file until after the last open
+        // handle is closed, and this protects us from race conditions where
+        // other processes replace the file out from underneath us.
+        drop(file);
+
+        Ok(())
     }
 }
 
 #[cfg(all(feature = "async_std", feature = "fs_utf8"))]
-impl DirExtUtf8 for cap_async_std::fs_utf8::Dir {
+#[async_trait]
+impl AsyncDirExtUtf8 for cap_async_std::fs_utf8::Dir {
     #[inline]
-    fn set_atime<P: AsRef<str>>(&self, path: P, atime: SystemTimeSpec) -> io::Result<()> {
+    async fn set_atime<P: AsRef<str> + Send>(
+        &self,
+        path: P,
+        atime: SystemTimeSpec,
+    ) -> io::Result<()> {
         let path = from_utf8(path)?;
-        set_times(
-            &self.as_filelike_view::<std::fs::File>(),
-            &path,
-            Some(atime),
-            None,
-        )
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times(
+                &clone.as_filelike_view::<std::fs::File>(),
+                &path,
+                Some(atime),
+                None,
+            )
+        })
+        .await
     }
 
     #[inline]
-    fn set_mtime<P: AsRef<str>>(&self, path: P, mtime: SystemTimeSpec) -> io::Result<()> {
+    async fn set_mtime<P: AsRef<str> + Send>(
+        &self,
+        path: P,
+        mtime: SystemTimeSpec,
+    ) -> io::Result<()> {
         let path = from_utf8(path)?;
-        set_times(
-            &self.as_filelike_view::<std::fs::File>(),
-            &path,
-            None,
-            Some(mtime),
-        )
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times(
+                &clone.as_filelike_view::<std::fs::File>(),
+                &path,
+                None,
+                Some(mtime),
+            )
+        })
+        .await
     }
 
     #[inline]
-    fn set_times<P: AsRef<str>>(
+    async fn set_times<P: AsRef<str> + Send>(
         &self,
         path: P,
         atime: Option<SystemTimeSpec>,
         mtime: Option<SystemTimeSpec>,
     ) -> io::Result<()> {
         let path = from_utf8(path)?;
-        set_times(
-            &self.as_filelike_view::<std::fs::File>(),
-            &path,
-            atime,
-            mtime,
-        )
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times(
+                &clone.as_filelike_view::<std::fs::File>(),
+                &path,
+                atime,
+                mtime,
+            )
+        })
+        .await
     }
 
     #[inline]
-    fn set_symlink_times<P: AsRef<str>>(
+    async fn set_symlink_times<P: AsRef<str> + Send>(
         &self,
         path: P,
         atime: Option<SystemTimeSpec>,
         mtime: Option<SystemTimeSpec>,
     ) -> io::Result<()> {
         let path = from_utf8(path)?;
-        set_times_nofollow(
-            &self.as_filelike_view::<std::fs::File>(),
-            &path,
-            atime,
-            mtime,
-        )
+        let clone = self.clone();
+        spawn_blocking(move || {
+            set_times_nofollow(
+                &clone.as_filelike_view::<std::fs::File>(),
+                &path,
+                atime,
+                mtime,
+            )
+        })
+        .await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn symlink<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    async fn symlink<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
-        symlink(&src, &self.as_filelike_view::<std::fs::File>(), &dst)
+        let clone = self.clone();
+        spawn_blocking(move || symlink(&src, &clone.as_filelike_view::<std::fs::File>(), &dst))
+            .await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn symlink_file<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
-        self.symlink(src, dst)
+    async fn symlink_file<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        self.symlink(src, dst).await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn symlink_dir<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
-        self.symlink(src, dst)
+    async fn symlink_dir<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        self.symlink(src, dst).await
     }
 
     #[cfg(windows)]
     #[inline]
-    fn symlink<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
-        if self.metadata(src.as_ref())?.is_dir() {
-            self.symlink_dir(src, dst)
+    async fn symlink<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
+        let src = from_utf8(src)?;
+        let src_ = src.clone();
+        let dst = from_utf8(dst)?;
+        let clone = self.clone();
+        // Call `stat` directly to avoid `async_trait` capturing `self`.
+        let metadata = spawn_blocking(move || {
+            stat(
+                &clone.as_filelike_view::<std::fs::File>(),
+                &src_,
+                FollowSymlinks::Yes,
+            )
+        })
+        .await?;
+        let clone = self.clone();
+        if metadata.is_dir() {
+            spawn_blocking(move || {
+                symlink_dir(&src, &clone.as_filelike_view::<std::fs::File>(), &dst)
+            })
+            .await
         } else {
-            self.symlink_file(src, dst)
+            spawn_blocking(move || {
+                symlink_file(&src, &clone.as_filelike_view::<std::fs::File>(), &dst)
+            })
+            .await
         }
     }
 
     #[cfg(windows)]
     #[inline]
-    fn symlink_file<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    async fn symlink_file<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
-        symlink_file(&src, &self.as_filelike_view::<std::fs::File>(), &dst)
+        let clone = self.clone();
+        spawn_blocking(move || symlink_file(&src, &clone.as_filelike_view::<std::fs::File>(), &dst))
+            .await
     }
 
     #[cfg(windows)]
     #[inline]
-    fn symlink_dir<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    async fn symlink_dir<P: AsRef<str> + Send, Q: AsRef<str> + Send>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
-        symlink_dir(&src, &self.as_filelike_view::<std::fs::File>(), &dst)
+        let clone = self.clone();
+        spawn_blocking(move || symlink_dir(&src, &clone.as_filelike_view::<std::fs::File>(), &dst))
+            .await
     }
 
     #[inline]
-    fn open_dir_nofollow<P: AsRef<str>>(&self, path: P) -> io::Result<Self> {
-        match open_dir_nofollow(
-            &self.as_filelike_view::<std::fs::File>(),
-            path.as_ref().as_ref(),
-        ) {
-            Ok(file) => Ok(Self::from_std_file(file.into(), ambient_authority())),
-            Err(e) => Err(e),
-        }
+    async fn open_dir_nofollow<P: AsRef<str> + Send>(&self, path: P) -> io::Result<Self> {
+        let path = from_utf8(path)?;
+        let clone = self.clone();
+        spawn_blocking(move || {
+            match open_dir_nofollow(&clone.as_filelike_view::<std::fs::File>(), path.as_ref()) {
+                Ok(file) => Ok(Self::from_std_file(file.into(), ambient_authority())),
+                Err(e) => Err(e),
+            }
+        })
+        .await
     }
 
     #[cfg(not(windows))]
     #[inline]
-    fn remove_file_or_symlink<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
-        self.remove_file(path.as_ref())
+    async fn remove_file_or_symlink<P: AsRef<str> + Send>(&self, path: P) -> io::Result<()> {
+        self.remove_file(path).await
     }
 
     #[cfg(windows)]
     #[inline]
-    fn remove_file_or_symlink<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
-        // This operation may race, because it checks the metadata before deleting
-        // the symlink. We tried to do this atomically by ReOpenFile with
-        // DELETE_ON_CLOSE but could not get it to work.
-        fn delete_symlink_to_dir(dir: &cap_async_std::fs_utf8::Dir, path: &str) -> io::Result<()> {
-            use crate::{FollowSymlinks, OpenOptionsFollowExt};
-            use cap_std::fs::OpenOptions;
-            use std::os::windows::fs::OpenOptionsExt;
-            use winapi::um::winbase::{FILE_FLAG_BACKUP_SEMANTICS, FILE_FLAG_OPEN_REPARSE_POINT};
+    async fn remove_file_or_symlink<P: AsRef<str> + Send>(&self, path: P) -> io::Result<()> {
+        use crate::{FollowSymlinks, OpenOptionsFollowExt};
+        use cap_primitives::fs::_WindowsByHandle;
+        use cap_std::fs::OpenOptions;
+        use std::os::windows::fs::OpenOptionsExt;
+        use winapi::um::winbase::{FILE_FLAG_BACKUP_SEMANTICS, FILE_FLAG_OPEN_REPARSE_POINT};
+        use winapi::um::winnt::{DELETE, FILE_ATTRIBUTE_DIRECTORY};
+        let path = path.as_ref();
 
-            let mut opts = OpenOptions::new();
-            opts.read(true);
-            opts.custom_flags(FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS);
-            opts.follow(FollowSymlinks::No);
-            let file = dir.open_with(path, &opts)?;
+        let mut opts = OpenOptions::new();
+        opts.access_mode(DELETE);
+        opts.custom_flags(FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS);
+        opts.follow(FollowSymlinks::No);
+        let file = self.open_with(path, &opts).await?;
 
-            let meta = file.metadata()?;
-            if meta.file_type().is_symlink() {
-                drop(file);
-                // Symlinks that point to directories use the remove_dir, not remove_file,
-                // operation on windows:
-                dir.remove_dir(path)?;
-                Ok(())
-            } else {
-                Err(io::Error::from_raw_os_error(
-                    winapi::shared::winerror::ERROR_DIRECTORY as i32,
-                ))
-            }
+        let meta = file.metadata()?;
+        if meta.file_type().is_symlink()
+            && meta.file_attributes() & FILE_ATTRIBUTE_DIRECTORY == FILE_ATTRIBUTE_DIRECTORY
+        {
+            self.remove_dir(path).await?;
+        } else {
+            self.remove_file(path).await?;
         }
 
-        self.remove_file(path.as_ref())
-            .or_else(|e| delete_symlink_to_dir(&self, path.as_ref()).map_err(|_| e))
+        // Drop the file after calling `remove_file` or `remove_dir`, since
+        // Windows doesn't actually remove the file until after the last open
+        // handle is closed, and this protects us from race conditions where
+        // other processes replace the file out from underneath us.
+        drop(file);
+
+        Ok(())
     }
 }
 
